import codecs


def to_bytes(s):
	"""
	If you don't know if a stringlike object is a Python 2 string (bytes) or Python 3 string (unicode).
	You could have a generic converter.
	Source: https://stackoverflow.com/questions/60368956/attributeerrorbytes-object-has-no-attribute-encode
	:type s: str or bytes
	:rtype: bytes
	"""
	if type(s) is bytes:
		return s
	elif type(s) is str:
		return codecs.encode(s, 'utf-8')
	else:
		raise TypeError(f"Expected bytes or string, but got {type(s)}")



import hashlib
import base64
import base32hex
from .make_hashable import represent_and_encode


def hash_object(obj, base=64):
	"""
	:type obj: Obj
	:type base: int
	:rtype str
	"""
	if hasattr(obj, '__hash64__') and base == 64:
		return obj.__hash64__()
	elif hasattr(obj, '__hash32__') and base == 32:
		return obj.__hash32__()

	hash_maker = hashlib.sha256()
	hash_maker.update(represent_and_encode(obj))
	if base == 64:
		return base64.b64encode(hash_maker.digest()).decode()
	elif base == 32:
		return base32hex.b32encode(hash_maker.digest()).replace('=', '-').replace('----', 'hb32')
	else:
		raise ValueError(f'base{base} is unknown!')


smart_hash = hash_object

hash = hash_object



from .hash_object import hash_object
from .hash_object import hash
from .hash_object import smart_hash
from .make_hashable import make_hashable
from .to_bytes import to_bytes

from .make_hashable import concat_pandas_columns, concat_spark_columns
from .make_hashable import map_pandas_to_int, map_spark_to_int
from .make_hashable import aggregate_pandas_int, aggregate_spark_int
from .make_hashable import represent_and_encode
from .memoize import memoize
from .Cache import Cache



import hashlib
import base64


def hash_file(path, base=64, block_size=65536):
	hash_maker = hashlib.sha256()
	with open(path, 'rb') as file:
		block = file.read(block_size)
		while len(block) > 0:
			hash_maker.update(block)
			block = file.read(block_size)
	if base == 64:
		return base64.b64encode(hash_maker.digest()).decode()
	elif base == 32:
		return base32hex.b32encode(hash_maker.digest()).replace('=', '-')
	else:
		raise ValueError(f'base{base} is unknown!')



from pyspark.sql.dataframe import DataFrame as SparkDF
from pandas import DataFrame as PandasDF

import numpy as np
from functools import reduce
from pyspark.sql.types import LongType
from pyspark.sql import functions as f
import hashlib
import base64

from .to_bytes import to_bytes


def hash_int(x):
	s = represent_and_encode(x)
	_hash_maker = hashlib.sha256()

	encoded = to_bytes(s)  # instead of x.encode()
	_hash_maker.update(encoded)
	return hash(base64.b64encode(_hash_maker.digest()).decode())


def concat_spark_columns(data, name='concat_column', sep='|', return_data=False):
	"""
	concatenates each row and creates a new column
	:type data: SparkDF
	:type name: str
	:type sep: str
	:type return_data: bool
	:param return_data: if True, the whole data frome with the new column is returned, otherwise just the column
	:rtype: SparkDF
	"""
	data = data.withColumn(
		name, f.concat_ws(
			sep, *[f.col(col) for col in data.columns]
		)
	)

	if return_data:
		return data
	else:
		return data.select(name)


def map_spark_to_int(data):
	def _map_function(row):
		"""
		:type row: Row
		:rtype: list[int]
		"""
		return [hash_int(x) for column, x in row.asDict().items()]

	return data.rdd.map(_map_function).toDF(data.columns)


def aggregate_spark_int(data):
	@f.udf(LongType())
	def spark_bitwise_xor(numbers):
		return reduce(lambda x, y: x ^ y, numbers)

	return data.agg(*[
		spark_bitwise_xor(f.collect_list(col)).alias(col)
		for col in data.columns
	])


def concat_pandas_columns(data, name='concat_column', sep='|', return_data=False):
	"""
	concatenates each row and creates a new column
	:type data: PandasDF
	:type name: str
	:type sep: str
	:type return_data: bool
	:param return_data: if True, the whole dataframe with the new column is returned, otherwise just the column
	:rtype: PandasDF
	"""
	def concat_row(x):
		return sep.join([str(value) for value in x.values if value is not None])

	data = data.copy()
	if name in data.columns:
		data.drop(columns=name, inplace=True)
	data[name] = data.apply(concat_row, axis=1)

	if return_data:
		return data
	else:
		return data[[name]]


def map_pandas_to_int(data):
	try:
		pandas_int_data = data.applymap(hash_int, na_action=None)
	except TypeError:
		pandas_int_data = data.apply(lambda x: x.map(hash_int, na_action=None))
	return pandas_int_data


def aggregate_pandas_int(data):
	def pandas_bitwise_xor(numbers):
		return reduce(lambda x, y: np.bitwise_xor(x, y), numbers)

	return data.agg(pandas_bitwise_xor, axis='rows')


def represent_and_encode(obj):
	class_representation = f'{obj.__class__.__name__}'
	obj_representation = repr(make_hashable(obj))
	representation = f'{class_representation}|{obj_representation}'
	return representation.encode()


def make_hashable(obj):
	if hasattr(obj, '__hashkey__'):
		return make_hashable(obj.__hashkey__())
	else:
		if isinstance(obj, (tuple, list)):
			return tuple((make_hashable(e) for e in obj))

		if isinstance(obj, dict):
			return tuple(sorted((k, make_hashable(v)) for k, v in obj.items()))

		if isinstance(obj, (set, frozenset)):
			return tuple(sorted(make_hashable(e) for e in obj))

		if isinstance(obj, (SparkDF, PandasDF)):
			return make_dataframe_hashable(data=obj)

		return obj


def make_dataframe_hashable(data):
	"""
	:type data: SparkDF or PandasDF
	:rtype: tuple
	"""

	# turn the dataframe into a single column of concatenated values
	# this makes sure if two cells in the same column are swapped the change is captured in this column
	# even though it will be ignored by the aggregation of each column
	data = data
	name = 'concat_column'
	sep = '|'
	return_data = True

	if isinstance(data, SparkDF):
		spark_compressed = concat_spark_columns(data=data, name=name, sep=sep, return_data=return_data)
		use = 'spark'
	elif isinstance(data, PandasDF):
		pandas_compressed = concat_pandas_columns(data=data, name=name, sep=sep, return_data=return_data)
		use = 'pandas'
	else:
		raise TypeError(f'data of type "{type(data)}" is not supported!')

	# convert each value in the column into an int hash

	if use == 'spark':
		spark_int_data = map_spark_to_int(spark_compressed)
		spark_aggregate = aggregate_spark_int(spark_int_data)
		aggregate_dictionary = spark_aggregate.first().asDict()

	elif use == 'pandas':
		pandas_int_data = map_pandas_to_int(pandas_compressed)
		pandas_aggregate = aggregate_pandas_int(pandas_int_data)
		aggregate_dictionary = pandas_aggregate.to_dict()

	else:
		raise RuntimeError(f'Usage "{use}" is not supported!')

	return tuple(sorted((k, v) for k, v in aggregate_dictionary.items()))



from disk import Path
from .hash_object import smart_hash


class Cache:
	def __init__(self, path, spark=None, echo=1):
		"""
		:type path: amazonian.S3.S3Path or Path or str
		:type spark: pyspark.sql.session.SparkSession
		"""
		if isinstance(path, str):
			path = Path(path)

		self._path = path
		self._spark = spark
		self._items = set()
		self._echo = echo

	@staticmethod
	def get_key(item):
		return smart_hash(item, base=32)

	def get_path(self, key):
		"""
		:type key: str
		:rtype: amazonian.S3.S3Path or Path
		"""
		return self._path / key

	def set_item(self, item, value):
		key = self.get_key(item)
		path = self.get_path(key)
		result = path.save(obj=value, mode='overwrite')
		if self._echo:
			print(f'{result} saved!')
		self._items.add(key)
		return result

	def get_item(self, item):
		key = self.get_key(item)
		path = self.get_path(key)
		if path.exists():
			if self._echo:
				print(f'{path.path} loaded!')
			return path.load(spark=self._spark)
		else:
			pickle_path = path + '.pickle'
			if pickle_path.exists():
				if self._echo:
					print(f'{pickle_path.path} loaded!')
				return pickle_path.load()
			else:
				parquet_path = path + '.parquet'
				if parquet_path.exists():
					if self._echo:
						print(f'{parquet_path.path} loaded!')
					return parquet_path.load(spark=self._spark)
				else:
					raise KeyError(f'key {key} does not exist!')

	def contains(self, item):
		key = self.get_key(item)
		if key in self._items:
			return True
		else:
			path = self.get_path(key)
			if path.exists():
				self._items.add(key)
				return True
			elif (path + '.pickle').exists():
				return True
			elif (path + '.parquet').exists():
				return True
			else:
				return False

	def delete_item(self, item):
		key = self.get_key(item)
		if key in self._items:
			self._items.remove(key)
			path = self.get_path(key)
			if path.exists():
				path.delete()
			else:
				raise FileNotFoundError(f'key "{key}" not found!')
		else:
			path = self.get_path(key)
			if path.exists():
				path.delete()
			else:
				raise KeyError(f'key "{key}" not found and path "{path}" not found!')

	def __setitem__(self, item, value):
		return self.set_item(item=item, value=value)

	def __getitem__(self, item):
		return self.get_item(item)

	def __contains__(self, item):
		return self.contains(item)

	def __delitem__(self, item):
		return self.delete_item(item)



import functools
from disk import Path
from .Cache import Cache


def memoize(path, exclude_kwargs=None, s3=None, spark=None, id=0):
	"""
	makes a decorator for memoizing a function
	:type path: amazonian.S3.S3Path or Path or str
	:type id: int
	:type exclude_kwargs: list or tuple
	:type s3: amazonian.S3.S3
	:type spark: pyspark.sql.session.SparkSession
	:rtype: callable
	"""

	if s3 is not None:
		if isinstance(path, str):
			path = s3 / path

	if spark is not None:
		try:
			if path.spark is None:
				path.set_spark(spark)
		except AttributeError:
			pass

	def decorator(function):
		"""
		memoizes a function
		:type function: callable
		:rtype: callable
		"""
		cache = Cache(path=path, spark=spark)

		if not isinstance(exclude_kwargs, (list, tuple)) and exclude_kwargs is not None:
			raise TypeError(f'exclude_kwargs is of type {type(exclude_kwargs)}! None, list, and tuple are accepted.')

		@functools.wraps(function)
		def wrapper(*args, **kwargs):
			if exclude_kwargs is not None:
				kwargs_in_key = {key: value for key, value in kwargs.items() if key not in exclude_kwargs}
			else:
				kwargs_in_key = kwargs.copy()

			key = (id, function.__name__, function.__doc__, args, kwargs_in_key)
			if key in cache:
				result = cache[key]
			else:
				result = function(*args, **kwargs)
				cache[key] = result
			return result

		wrapper.cache = cache
		wrapper.excluded_kwargs = exclude_kwargs
		return wrapper
	return decorator
